# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Entrypoint of the package"""

import json
from pathlib import Path
from typing import Optional

import typer
import yaml
from linkml_runtime.utils.schemaview import SchemaView

from ghga_validator.core.models import ValidationReport
from ghga_validator.core.validator import Validator
from ghga_validator.plugins.base_plugin import ValidationPlugin
from ghga_validator.plugins.utils import discover_plugins
from ghga_validator.schema_utils import get_target_class

cli = typer.Typer()

DEFAULT_PLUGINS = ["GHGAJsonSchemaValidationPlugin"]

VALIDATION_PLUGINS = ["RefValidationPlugin", "UniqueIdentifierValidationPlugin"]


def validate_json_file(
    file: Path, schema: Path, report: Path, target_class: str
) -> bool:
    """
    Validate JSON object read from a file against a given schema.
    Store the errors to the validation report.
    Args:
        file: The URL or path to file containing data to be validated
        schema: The URL or path to YAML file
        report: The URL or path to store the validation results
    """
    with open(file, encoding="utf8") as json_file:
        submission_json = yaml.safe_load(json_file)
    if submission_json is None:
        raise EOFError(f"<{file}> is empty! Nothing to validate!")
    schema_view = SchemaView(schema)
    validation_report = validate(
        schema_view,
        target_class=target_class,
        data=submission_json,
        plugins=load_plugins(DEFAULT_PLUGINS, schema_view),
    )
    if validation_report.valid:
        default_validation_results = validation_report.validation_results
        validation_report = validate(
            schema_view,
            target_class=target_class,
            data=submission_json,
            plugins=load_plugins(VALIDATION_PLUGINS, schema_view),
        )
        validation_report.validation_results = (
            default_validation_results + validation_report.validation_results
        )
    else:
        typer.echo(
            "JSON schema validation failed. Subsequent validations skipped.", err=True
        )

    with open(report, "w", encoding="utf-8") as sub:
        json.dump(
            validation_report.dict(
                exclude={"object"}, exclude_unset=True, exclude_none=True
            ),
            sub,
            ensure_ascii=False,
            indent=4,
        )
    return validation_report.valid


def validate(
    schema: SchemaView,
    target_class: str,
    data: dict,
    plugins: list,
) -> ValidationReport:
    """
    Validate an object of a particular type against a given schema.
    Args:
        schema: Virtual LinkML schema (SchemaView)
        target_class: The root class name
        data: The JSON object to validate
        plugins: List of plugin class names for validation
    """
    validator = Validator(schema=schema, plugins=plugins)
    report = validator.validate(data, target_class)
    return report


def load_plugins(plugin_types: list[str], schema: SchemaView) -> list[ValidationPlugin]:
    """Load the list of plugins"""
    plugin_list = []
    discovered_plugins = discover_plugins(ValidationPlugin)
    for plugin_name in plugin_types:
        if plugin_name in discovered_plugins:
            plugin_class = discovered_plugins[plugin_name]
            plugin_list.append(plugin_class(schema=schema))
        else:
            raise ModuleNotFoundError(f"Plugin '{plugin_name}' not found")
    return plugin_list


@cli.command()
def main(
    schema: Path = typer.Option(
        ..., "--schema", "-s", help="Path to metadata schema (modelled using LinkML)"
    ),
    input_file: Path = typer.Option(
        ...,
        "--input",
        "-i",
        file_okay=True,
        dir_okay=False,
        help="Path to submission file in JSON format to be validated",
    ),
    report: Path = typer.Option(
        ...,
        "--report",
        "-r",
        file_okay=True,
        dir_okay=False,
        writable=True,
        help="Path to resulting validation report",
    ),
    target_class: Optional[str] = typer.Option(None, help="The root class name"),
):
    """
    GHGA Validator

    ghga-validator is a command line utility to validate metadata w.r.t. its
    compliance to the GHGA Metadata Model. It takes metadata encoded in JSON of
    YAML format and produces a validation report in JSON format.
    """
    typer.echo("Start validating...")
    if not target_class:
        target_class = get_target_class(str(Path(schema).resolve()))
    if not target_class:
        raise TypeError(
            "Target class cannot be inferred,"
            + "please specify the 'target_class' argument"
        )
    if validate_json_file(input_file, schema, report, target_class):
        typer.echo(f"<{input_file}> is valid!")
    else:
        typer.echo(
            f"<{input_file}> is invalid! Look at <{report}> for validation report"
        )



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utils"""


def path_as_string(error_path: list) -> str:
    """Convert the path to the error in JSON to string format
    Args:
        error_path (List): path to the error in JSON
    Returns:
        str: string representation of the error path
    """
    path_str = ".".join(str(elem) for elem in error_path)
    return path_str



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Provides an ObjectIterator for LinkML data."""

from collections.abc import Iterable, Iterator
from copy import deepcopy
from itertools import chain
from numbers import Number
from typing import Optional, Union

from linkml_runtime.linkml_model.meta import ClassDefinitionName, SlotDefinition
from linkml_runtime.utils.schemaview import SchemaView


class RootInferenceError(RuntimeError):
    """This error is produced if the root class could not be determined from a
    LinkML schema.
    """


class ObjectIterator:
    """This iterator class enables iterating through all elements below a
    specified or inferred root element. The iterator returns tuples of an
    elements class name, identifier if present and the corresponding element
    data, which has been re-serialized such that all identifiable inlined
    elements below the element itself have been un-inlined, i.e. replaced by
    their identifiers.
    """

    _schema: SchemaView
    _root: Union[ClassDefinitionName, str]
    _data: dict
    _recursion_iterator: Optional[Iterator]
    _enumerate_non_identifiable: bool
    _inline_non_identifiable: bool
    _path: list

    def __init__(  # noqa: PLR0913
        self,
        schema: SchemaView,
        data: dict,
        root: Optional[str] = None,
        enumerate_non_identifiable=False,
        inline_non_identifiable=True,
        path: Optional[list] = None,
    ):  # pylint: disable=too-many-arguments
        """Creates a new IdentifiedObjectIterator."""
        self._schema = schema
        self._data = data
        self._enumerate_non_identifiable = enumerate_non_identifiable
        self._inline_non_identifiable = inline_non_identifiable
        self._path = path if path else []
        # If a root class was specified, use it
        if root:
            self._root = root
        # ... otherwise, attempt to infer the root class from the provided model.
        else:
            self._root = ObjectIterator._infer_root(schema)

        # Root class slots that we need to recurse into, i.e. all slots with a
        # class range.
        self._recursion_slots = [
            (slot_def.name, slot_def)
            for slot_def in self._schema.class_induced_slots(self._root)
            if slot_def.range in schema.all_classes() and slot_def.inlined is not False
        ]

        self._recursion_iterator = None

    @staticmethod
    def _infer_root(schema: SchemaView) -> ClassDefinitionName:
        """Iterates through all class definitions in a schema and returns the
        name of the single class that is defined as the tree root.

        Raises a RootInferenceError if no or multiple such classes are found.
        """
        # Identify all classes that have tree_root set to true
        root_labeled_classes = [
            name
            for name, definition in schema.all_classes().items()
            if definition.tree_root
        ]

        # If there are no or multiple classes, raise an error
        if not root_labeled_classes:
            raise RootInferenceError("Schema tree root class not found.")
        if len(root_labeled_classes) > 1:
            raise RootInferenceError("Schema has multiple classes defined as tree root")

        return root_labeled_classes[0]

    @staticmethod
    def _re_serialize_element(
        data: dict,
        schema: SchemaView,
        root: Union[str, ClassDefinitionName],
        inline_non_identifiable: bool,
    ):
        """Re-serializes the element serialized in the passed data such that all
        inlined slots are not inlined anymore if the corresponding slot class
        itself has an identifier slot.
        """
        re_serialized_element = {}
        for slot_name, slot_value in data.items():
            slot_def = schema.induced_slot(slot_name, root)
            # If the slot has an inlined class range, transform the inlined
            # value into a reference if it has an identifier slot
            if slot_def.range in schema.all_classes() and slot_def.inlined is not False:
                id_slot: Optional[SlotDefinition] = schema.get_identifier_slot(
                    slot_def.range
                )
                # If the slot class has no identifier, recursively serialize it
                # if enabled
                if id_slot is None and inline_non_identifiable:
                    if slot_def.multivalued:
                        # Sanity check that the slot value is a list
                        if not isinstance(slot_value, list):
                            raise RuntimeError(
                                "Multivalued, inlined slot values must be"
                                " inlined as lists if the slot range itself does"
                                " not have an identifier slot."
                            )
                        re_serialized_element[slot_name] = [
                            ObjectIterator._re_serialize_element(
                                v, schema, slot_def.range, inline_non_identifiable
                            )
                            for v in slot_value
                        ]
                    else:
                        re_serialized_element[
                            slot_name
                        ] = ObjectIterator._re_serialize_element(
                            slot_value,
                            schema,
                            slot_def.range,
                            inline_non_identifiable,
                        )
                elif id_slot is not None:  # noqa: SIM102
                    # If the slot is multivalued, we must differentiate whether it
                    # is inlined as list or as dictionary
                    if slot_def.multivalued:
                        # The data is of type list and this is permitted
                        # according to the slot definition
                        if (
                            slot_def.inlined_as_list or slot_def.inlined_as_list is None
                        ) and isinstance(slot_value, list):
                            re_serialized_element[slot_name] = [
                                elem[id_slot.name] for elem in slot_value
                            ]
                        # The data is of type dict and this is permitted
                        # according to the slot definition
                        elif (
                            slot_def.inlined_as_list is False
                            or slot_def.inlined_as_list is None
                        ) and isinstance(slot_value, dict):
                            re_serialized_element[slot_name] = list(slot_value)
                        # The data is of a type that is not permitted according
                        # to the slot definition
                        else:
                            raise RuntimeError(
                                "Invalid data. Slot is configured as"
                                f" multivalued={slot_def.multivalued} but data is of"
                                f" type {type(slot_value).__name__}"
                            )

            # If the slot is an enum, type or a non-inlined class, keep the
            # value as it is
            else:
                re_serialized_element[slot_name] = slot_value

        return re_serialized_element

    def _re_serialize_root(self):
        """Re-serializes the data stored in this iterator object such that
        inlined slots are not inlined anymore if the corresponding slot class
        itself has an identifier slot.
        """
        return ObjectIterator._re_serialize_element(
            self._data, self._schema, self._root, self._inline_non_identifiable
        )

    def _child_iterators(self) -> Iterable[Iterator]:
        """Returns an iterable of IdentifiedObjectIterator objects for every
        class-range slot of the current root class that has not been iterated
        yet.
        """
        for next_slot_name, next_slot_def in self._recursion_slots:
            if next_slot_name in self._data.keys():  # noqa: SIM118
                # If the slot is not multivalued, a single-value list is returned
                # with an IdentifiedObjectIterator for the value of the slot
                if not next_slot_def.multivalued:
                    yield ObjectIterator(
                        self._schema,
                        self._data[next_slot_name],
                        next_slot_def.range,
                        enumerate_non_identifiable=self._enumerate_non_identifiable,
                        inline_non_identifiable=self._inline_non_identifiable,
                        path=self._path + [next_slot_name],  # noqa: RUF005
                    )
                # If the slot is multivalued and encoded in list format, a list with
                # one IdentifiedObjectIterator per element is returned
                elif (
                    next_slot_def.inlined_as_list
                    or next_slot_def.inlined_as_list is None
                ) and isinstance(self._data[next_slot_name], list):
                    for idx, elem in enumerate(self._data[next_slot_name]):
                        yield ObjectIterator(
                            self._schema,
                            elem,
                            next_slot_def.range,
                            enumerate_non_identifiable=self._enumerate_non_identifiable,
                            inline_non_identifiable=self._inline_non_identifiable,
                            path=self._path + [next_slot_name] + [idx],  # noqa: RUF005
                        )
                # If the slot is multivalued and encoded in dictionary format, a list with
                # one IdentifiedObjectIterator per element is returned. Since the
                # identifier slot is optional in the dictionary format, the
                # identifier is set based on the dictionary keys before the data is
                # used.
                elif (
                    next_slot_def.inlined_as_list is False
                    or next_slot_def.inlined_as_list is None
                ) and isinstance(self._data[next_slot_name], dict):
                    identifier_slot = self._schema.get_identifier_slot(
                        next_slot_def.range
                    )
                    if identifier_slot is None:
                        raise RuntimeError(
                            f"Expected identifier slot for {next_slot_def.range}"
                        )
                    modified_data = deepcopy(self._data[next_slot_name])
                    for key, value in modified_data.items():
                        value[identifier_slot.name] = key
                    for key, elem in modified_data.items():
                        yield ObjectIterator(
                            self._schema,
                            elem,
                            next_slot_def.range,
                            enumerate_non_identifiable=self._enumerate_non_identifiable,
                            inline_non_identifiable=self._inline_non_identifiable,
                            path=self._path + [next_slot_name] + [key],  # noqa: RUF005
                        )
                # If none of the previous conditions were met, we have encountered a
                # data format that is incompatible with the multivalued, inlined and
                # inlined_as_list configurations.
                else:
                    raise RuntimeError(
                        "Invalid data. Slot is configured as"
                        f" inlined_as_list={next_slot_def.inlined_as_list} but data is"
                        f" of type {type(self._data[next_slot_name]).__name__}"
                    )

    def __next__(
        self,
    ) -> tuple[
        Union[str, ClassDefinitionName],
        Optional[Union[str, Number]],
        dict,
        list[Union[str, Number]],
    ]:
        """Select the next element"""
        if self._recursion_iterator is None:
            # Build an iterator for all slots of the root class that have a class range
            self._recursion_iterator = chain.from_iterable(self._child_iterators())

            # De-serialize the root element if it is identifiable
            root_identifier_slot = self._schema.get_identifier_slot(self._root)
            if root_identifier_slot or self._enumerate_non_identifiable:
                return (
                    self._root,  # root element class
                    self._data[root_identifier_slot.name]
                    if root_identifier_slot
                    else None,  # root element identifier
                    self._re_serialize_root(),  # root element data
                    self._path,
                )

        # Pick the next slot with a class range and recurse
        return next(self._recursion_iterator)

    def __iter__(self):
        """Returns the iterator itself."""
        return self



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test data validation (using all validation plugins)"""

import os

from ghga_validator.cli import validate_json_file

from .fixtures.utils import BASE_DIR


def test_validate_slots():
    """Test data validation (using all validation plugins)"""
    schema = BASE_DIR / "schemas" / "example_schema.yaml"
    file = BASE_DIR / "data" / "example_data.json"
    report = BASE_DIR / "tmp.json"
    target_class = "TextAnalysis"

    assert validate_json_file(file, schema, report, str(target_class)) is True
    if os.path.exists(report):
        os.remove(report)



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plugin for validating the identifier uniqueness"""

from ghga_validator.core.models import ValidationMessage, ValidationResult
from ghga_validator.linkml.object_iterator import ObjectIterator
from ghga_validator.plugins.base_plugin import ValidationPlugin
from ghga_validator.utils import path_as_string


class UniqueIdentifierValidationPlugin(ValidationPlugin):
    """
    Plugin to check whether the fields defined as identifier/unique key
    are unique for a class.
    """

    NAME = "UniqueIdentifierValidationPlugin"

    def validate(self, data: dict, target_class: str) -> ValidationResult:
        """
        Perform validation on an object.

        Args:
            data: The JSON object to validate
            target_class: class name for root class

        Returns:
            ValidationResult: A validation result that describes the outcome of validation

        """
        messages = self.validate_unique_fields(data, target_class)
        valid = len(messages) == 0

        result = ValidationResult(
            plugin_name=self.NAME, valid=valid, validation_messages=messages
        )
        return result

    def validate_unique_fields(
        self,
        object_to_validate: dict,
        target_class: str,
    ) -> list[ValidationMessage]:
        """
        Validate non inlined reference fields in a JSON object

        Args:
            object_to_validate: input JSON object
            target_class: parent class in the schema

        Returns:
            SlotDefinition: class definition

        """
        messages = []

        seen_ids: dict[tuple, list] = {}
        for class_name, identifier, data, path in ObjectIterator(
            self.schema, object_to_validate, target_class
        ):
            id_slot = self.schema.get_identifier_slot(class_name)
            id_slot_name = id_slot.name if id_slot is not None else "UNKNOWN"
            if (class_name, identifier) in seen_ids:
                previous_path = seen_ids[class_name, identifier]
                message = ValidationMessage(
                    message="Duplicate value for identifier, "
                    + f"same value used at {path_as_string(previous_path)}.",
                    field=f"{path_as_string(path)}.{id_slot_name}",
                    value=data[id_slot_name],
                )
                messages.append(message)
            else:
                seen_ids[class_name, identifier] = [*path, id_slot_name]
        return messages



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Validator of data against a given LinkML schema."""

from linkml_runtime.utils.schemaview import SchemaView

from ghga_validator.core.models import ValidationReport
from ghga_validator.plugins.base_plugin import ValidationPlugin


class Validator:
    """
    Validator of data against a given LinkML schema.

    Args:
        schema: Virtual LinkML schema (SchemaView)
        plugins: List of plugins for validation

    """

    def __init__(self, schema: SchemaView, plugins: list[ValidationPlugin]) -> None:
        self._schema = schema
        self._plugins = plugins

    def validate(self, data: dict, target_class: str) -> ValidationReport:
        """
        Validate an object.

        Args:
            data: The object to validate
            target_class: The type of object

        Returns:
            ValidationReport: A validation report that summarizes the validation

        """
        validation_results = [
            plugin.validate(data=data, target_class=target_class)
            for plugin in self._plugins
        ]
        all_valid = all(result.valid for result in validation_results)
        validation_report = ValidationReport(
            object=data,
            type=target_class,
            valid=all_valid,
            validation_results=validation_results,
        )
        return validation_report



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utils for LinkML schema"""

from typing import Optional

from linkml.utils.datautils import infer_root_class
from linkml_runtime.utils.schemaview import SchemaView, SlotDefinition


def get_range_class(schema_view, slot_def: SlotDefinition) -> Optional[str]:
    """Return the range class for a slot

    Args:
        slot_def (SlotDefinition): Slot Definition

    Returns:
        Optional[str]: Range class for a slot
    """
    return slot_def.range if slot_def.range in schema_view.all_classes() else None


def get_target_class(schema: str) -> Optional[str]:
    """
    Infer the root class from schema
    Args:
        schema (str): YAML schema as the string

    Returns:
        class name for root class, if found in the scheme
    """
    with open(schema, encoding="utf8") as file:
        input_schema = file.read()
        return infer_root_class(SchemaView(input_schema))



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Collection of classes for modelling validation results"""

from typing import Any, Optional

from pydantic import BaseModel


class ValidationMessage(BaseModel):
    """ValidationMessage represents a validation error message"""

    field: Optional[str] = None
    value: Optional[Any] = None
    message: str


class ValidationResult(BaseModel):
    """
    ValidationResult represents the results of validation
    by a plugin.
    """

    plugin_name: str
    valid: bool
    validation_messages: list[ValidationMessage] = []


class ValidationReport(BaseModel):
    """
    ValidationReport represents the overall validation result by all plugins
    for a given object.
    """

    object: Optional[dict]
    type: str
    valid: bool
    validation_results: list[ValidationResult]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plugin for structural validation of a JSON object"""

import json

import jsonschema
from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml_runtime.utils.schemaview import ClassDefinitionName

from ghga_validator.core.models import ValidationMessage, ValidationResult
from ghga_validator.plugins.base_plugin import ValidationPlugin
from ghga_validator.utils import path_as_string


class GHGAJsonSchemaValidationPlugin(ValidationPlugin):
    """Plugin for structural validation of a JSON object."""

    NAME = "GHGAJsonSchemaValidationPlugin"

    def validate(
        self, data: dict, target_class: ClassDefinitionName
    ) -> ValidationResult:
        """
        Perform validation on an object.

        Args:
            data: The JSON object to validate
            target_class: class name for root class

        Returns:
            ValidationResult: A validation result that describes the outcome of validation

        """
        json_schema = self.jsonschema_from_linkml(target_class)

        messages = []

        validator = jsonschema.Draft7Validator(json_schema)
        errors = validator.iter_errors(data)

        for error in errors:
            message = ValidationMessage(
                message=error.message,
                field=path_as_string(error.absolute_path),
                value=error.instance,
            )
            messages.append(message)
            for err in error.context:
                message = ValidationMessage(
                    message=err.message,
                    field=path_as_string(err.absolute_path),
                    value=err.instance,
                )
                messages.append(message)

        valid = len(messages) == 0

        result = ValidationResult(
            plugin_name=self.NAME, valid=valid, validation_messages=messages
        )
        return result

    def jsonschema_from_linkml(self, target_class: ClassDefinitionName) -> dict:
        """Generates JSON schema from a LinkML schema"""
        json_schema_as_string = JsonSchemaGenerator(
            schema=self.schema.schema, top_class=target_class
        ).serialize()
        json_schema = json.loads(json_schema_as_string)
        return json_schema



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plugin for validating the non inline references"""

from collections import defaultdict
from numbers import Number
from typing import Union

from ghga_validator.core.models import ValidationMessage, ValidationResult
from ghga_validator.linkml.object_iterator import ObjectIterator
from ghga_validator.plugins.base_plugin import ValidationPlugin
from ghga_validator.schema_utils import get_range_class
from ghga_validator.utils import path_as_string


# pylint: disable=too-many-locals
class RefValidationPlugin(ValidationPlugin):
    """
    Plugin to check whether the values in non inline reference fields point
    to existing objects.
    """

    NAME = "RefValidationPlugin"

    def validate(self, data: dict, target_class: str) -> ValidationResult:
        """
        Perform validation on an object.

        Args:
            data: The object to validate
            target_class: class name for root class

        Returns:
            ValidationResult: A validation result that describes the outcome of validation

        """
        all_class_ids = self.get_all_class_ids(data, target_class)
        messages = self.validate_refs(data, target_class, all_class_ids)

        valid = len(messages) == 0

        result = ValidationResult(
            plugin_name=self.NAME, valid=valid, validation_messages=messages
        )
        return result

    def get_all_class_ids(self, obj: dict, target_class: str) -> dict[str, list[str]]:
        """Get all lists of identifies of inlined objects organized by class name

        Args:
            obj (Dict): The object to be parsed
            target_class (str): Target class

        Returns:
            Dict[class_name, List[str]]: The dictionary containing the lists of
            identifiers by the class name
        """
        all_ids = defaultdict(list)

        for class_name, identifier, _, _ in ObjectIterator(
            self.schema, obj, target_class
        ):
            all_ids[class_name].append(identifier)

        return all_ids

    def validate_refs(
        self,
        object_to_validate: dict,
        target_class: str,
        all_class_ids: dict,
    ) -> list[ValidationMessage]:
        """
        Validate non inlined reference fields in the JSON data

        Args:
            object_to_validate: input data
            target_class: parent class in the schema
            all_class_ids: pre-computed dictionary containing all identifiers ordered by class

        Returns:
            List[ValidationMessage]: List of validation messages

        """
        messages = []

        for class_name, _, data, path in ObjectIterator(
            self.schema, object_to_validate, target_class
        ):
            for field, value in data.items():
                slot_def = self.schema.induced_slot(field, class_name)
                range_class = get_range_class(self.schema, slot_def)
                if range_class and not self.schema.is_inlined(slot_def):
                    non_match = self.find_missing_refs(
                        value, all_class_ids[range_class]
                    )
                    if len(non_match) == 0:
                        continue
                    message = ValidationMessage(
                        message="Unknown reference(s) " + str(non_match),
                        field=f"{path_as_string(path)}.{field}",
                        value=value,
                    )
                    messages.append(message)
        return messages

    def find_missing_refs(
        self,
        ref_value: Union[list[Union[Number, str]], Union[Number, str]],
        id_list: list,
    ) -> list:
        """
        Search for missing references

        Returns:
            List: List of missing references
        """
        if not isinstance(ref_value, list):
            return [ref_value] if ref_value not in id_list else []
        return [x for x in ref_value if x not in id_list]



# Copyright 2021 - 2023 Universität Tübingen, DKFZ, EMBL, and Universität zu Köln
# for the German Human Genome-Phenome Archive (GHGA)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Utils for plugins"""

import importlib
import pkgutil

import ghga_validator.plugins as plugin_package


def discover_plugins(plugin_type) -> dict:
    """Discover all plugins of a type"""
    discovered_plugins = {}
    for _, module_name, _ in pkgutil.iter_modules(plugin_package.__path__):
        try:
            module = importlib.import_module(f"{plugin_package.__name__}.{module_name}")
            for name, cls in module.__dict__.items():
                if isinstance(cls, type) and issubclass(cls, plugin_type):
                    discovered_plugins[name] = cls
        except ImportError as err:
            print(f"Error loading module '{module_name}': {err}")
    return discovered_plugins
